# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import array
import os

import numpy as np
import pytest

import uproot

ROOT = pytest.importorskip("ROOT")


def test_basic(tmp_path):
    # original = os.path.join(tmp_path, "original.root")
    newfile = os.path.join(tmp_path, "newfile.root")

    # f1 = ROOT.TFile(original, "recreate")
    # f1.SetCompressionLevel(0)
    # t1 = ROOT.TTree("t1", "title")
    # d1 = array.array("d", [0.0])
    # t1.Branch("branch1", d1, "branch1/D")

    # t1.Write()
    # f1.Close()

    # with uproot.open(original) as fin:
    #     fin["t1"]

    with uproot.recreate(newfile, compression=None) as fout:
        tree = fout.mktree("t1", {"branch1": np.float64}, "title")

    f2 = ROOT.TFile(newfile)
    t2 = f2.Get("t1")

    assert t2.GetName() == "t1"
    assert t2.GetTitle() == "title"

    assert t2.GetBranch("branch1").GetName() == "branch1"
    assert t2.GetBranch("branch1").GetTitle() == "branch1/D"

    assert t2.GetBranch("branch1").GetLeaf("branch1").GetName() == "branch1"
    assert t2.GetBranch("branch1").GetLeaf("branch1").GetTitle() == "branch1"

    assert t2.GetLeaf("branch1").GetName() == "branch1"
    assert t2.GetLeaf("branch1").GetTitle() == "branch1"

    f2.Close()


def test_rename(tmp_path):
    newfile = os.path.join(tmp_path, "newfiley_file.root")

    with uproot.recreate(newfile, compression=None) as fout:
        tree = fout.mktree(
            "treey_tree",
            {"branchy_branch": np.float64},
            "titley_title",
        )

    f2 = ROOT.TFile(newfile)
    t2 = f2.Get("treey_tree")

    assert t2.GetName() == "treey_tree"
    assert t2.GetTitle() == "titley_title"

    assert t2.GetBranch("branchy_branch").GetName() == "branchy_branch"
    assert t2.GetBranch("branchy_branch").GetTitle() == "branchy_branch/D"

    assert (
        t2.GetBranch("branchy_branch").GetLeaf("branchy_branch").GetName()
        == "branchy_branch"
    )
    assert (
        t2.GetBranch("branchy_branch").GetLeaf("branchy_branch").GetTitle()
        == "branchy_branch"
    )

    assert t2.GetLeaf("branchy_branch").GetName() == "branchy_branch"
    assert t2.GetLeaf("branchy_branch").GetTitle() == "branchy_branch"

    f2.Close()


def test_2_branches(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        tree = fout.mktree(
            "t1",
            {"branch1": np.float64, "branch2": np.float64},
            "title",
        )

    f2 = ROOT.TFile(newfile)
    t2 = f2.Get("t1")

    assert t2.GetName() == "t1"
    assert t2.GetTitle() == "title"

    for branchname in ["branch1", "branch2"]:
        assert t2.GetBranch(branchname).GetName() == branchname
        assert t2.GetBranch(branchname).GetTitle() == branchname + "/D"

        assert t2.GetBranch(branchname).GetLeaf(branchname).GetName() == branchname
        assert t2.GetBranch(branchname).GetLeaf(branchname).GetTitle() == branchname

        assert t2.GetLeaf(branchname).GetName() == branchname
        assert t2.GetLeaf(branchname).GetTitle() == branchname

    # FIXME: also test the following in a case where "t1" has a 64-bit TKey
    with uproot.open(newfile) as fin:
        t1 = fin["t1"]
        branch1 = t1["branch1"]
        branch2 = t1["branch2"]
        assert branch1.member("fLeaves")[0] is t1.member("fLeaves")[0]
        assert branch2.member("fLeaves")[0] is t1.member("fLeaves")[1]

    f2.Close()


def test_100_branches(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        tree = fout.mktree(
            "t1",
            {"branch" + str(i): np.float64 for i in range(100)},
            "title",
        )

    f2 = ROOT.TFile(newfile)
    t2 = f2.Get("t1")

    assert t2.GetName() == "t1"
    assert t2.GetTitle() == "title"

    for branchname in ["branch" + str(i) for i in range(100)]:
        assert t2.GetBranch(branchname).GetName() == branchname
        assert t2.GetBranch(branchname).GetTitle() == branchname + "/D"

        assert t2.GetBranch(branchname).GetLeaf(branchname).GetName() == branchname
        assert t2.GetBranch(branchname).GetLeaf(branchname).GetTitle() == branchname

        assert t2.GetLeaf(branchname).GetName() == branchname
        assert t2.GetLeaf(branchname).GetTitle() == branchname

    f2.Close()


def test_branch_types(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        tree = fout.mktree(
            "t1",
            {
                "typeO": np.bool_,
                "typeB": np.int8,
                "typeb": np.uint8,
                "typeS": np.int16,
                "types": np.uint16,
                "typeI": np.int32,
                "typei": np.uint32,
                "typeL": np.int64,
                "typel": np.uint64,
                "typeF": np.float32,
                "typeD": np.float64,
            },
            "title",
        )

    f2 = ROOT.TFile(newfile)
    t2 = f2.Get("t1")

    assert t2.GetName() == "t1"
    assert t2.GetTitle() == "title"

    for name, size, isunsigned in [
        ("typeO", 1, False),
        ("typeB", 1, False),
        ("typeb", 1, True),
        ("typeS", 2, False),
        ("types", 2, True),
        ("typeI", 4, False),
        ("typei", 4, True),
        ("typeL", 8, False),
        ("typel", 8, True),
        ("typeF", 4, False),
        ("typeD", 8, False),
    ]:
        assert t2.GetBranch(name).GetName() == name
        assert t2.GetBranch(name).GetTitle() == name + "/" + name[-1]

        assert t2.GetBranch(name).GetLeaf(name).GetName() == name
        assert t2.GetBranch(name).GetLeaf(name).GetTitle() == name

        assert t2.GetLeaf(name).GetName() == name
        assert t2.GetLeaf(name).GetTitle() == name
        assert t2.GetLeaf(name).GetLenType() == size
        assert t2.GetLeaf(name).IsUnsigned() == isunsigned

    f2.Close()


def test_basket(tmp_path):
    # original = os.path.join(tmp_path, "original.root")
    newfile = os.path.join(tmp_path, "newfile.root")

    # f1 = ROOT.TFile(original, "recreate")
    # f1.SetCompressionLevel(0)
    # t1 = ROOT.TTree("t1", "title")
    # d1 = array.array("i", [0])
    # t1.Branch("branch1", d1, "branch1/I")

    # d1[0] = 5
    # t1.Fill()
    # d1[0] = 4
    # t1.Fill()
    # d1[0] = 3
    # t1.Fill()
    # d1[0] = 2
    # t1.Fill()
    # d1[0] = 1
    # t1.Fill()
    # d1[0] = 5
    # t1.Fill()

    # t1.Write()
    # f1.Close()

    # with uproot.open(original) as fin:
    #     t1 = fin["t1"]

    with uproot.recreate(newfile, compression=None) as fout:
        tree = fout.mktree("t2", {"branch1": np.int32}, "title")
        tree.extend({"branch1": np.array([5, 4, 3, 2, 1, 5], ">i4")})

    with uproot.open(newfile) as fin2:
        t2 = fin2["t2"]
        assert t2["branch1"].array(library="np").tolist() == [5, 4, 3, 2, 1, 5]

    f2 = ROOT.TFile(newfile)
    t2 = f2.Get("t2")
    assert [x.branch1 for x in t2] == [5, 4, 3, 2, 1, 5]
    f2.Close()


def test_baskets_branches(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    b1 = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    b2 = [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]

    with uproot.recreate(newfile, compression=None) as fout:
        tree = fout.mktree("t", {"b1": np.int32, "b2": np.float64}, "title")
        tree.extend({"b1": b1, "b2": b2})
        tree.extend({"b1": b1, "b2": b2})

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("t")
    assert t1.GetEntries() == len(b1) + len(b1)
    assert [x.b1 for x in t1] == b1 + b1
    assert [x.b2 for x in t1] == b2 + b2

    with uproot.open(newfile + ":t") as t2:
        assert t2.num_entries == len(b1) + len(b1)
        assert t2["b1"].array(library="np").tolist() == b1 + b1
        assert t2["b2"].array(library="np").tolist() == b2 + b2

    f1.Close()


def test_baskets_beyond_capacity(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    b1 = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    b2 = [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]

    with uproot.recreate(newfile, compression=None) as fout:
        tree = fout.mktree("t", {"b1": np.int32, "b2": np.float64}, "title")

        assert tree._cascading._basket_capacity == 10

        for _ in range(5):
            fout["t"].extend({"b1": b1, "b2": b2})

        assert tree._cascading._basket_capacity == 10

        for _ in range(10):
            fout["t"].extend({"b1": b1, "b2": b2})

        assert tree._cascading._basket_capacity == 100

        for _ in range(90):
            fout["t"].extend({"b1": b1, "b2": b2})

        assert tree._cascading._basket_capacity == 1000

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("t")
    assert t1.GetEntries() == len(b1) * 105
    assert [x.b1 for x in t1] == b1 * 105
    assert [x.b2 for x in t1] == b2 * 105

    with uproot.open(newfile) as fin:
        assert fin.keys() == ["t;1"]  # same cycle number
        t2 = fin["t"]
        assert t2.num_entries == len(b1) * 105
        assert t2["b1"].array(library="np").tolist() == b1 * 105
        assert t2["b2"].array(library="np").tolist() == b2 * 105

    f1.Close()


def test_writable_vs_readable_tree(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    f1 = ROOT.TFile(newfile, "recreate")
    f1.SetCompressionLevel(0)
    t1 = ROOT.TTree("t1", "title")
    d1 = array.array("i", [0])
    t1.Branch("branch1", d1, "branch1/I")

    d1[0] = 5
    t1.Fill()
    d1[0] = 4
    t1.Fill()
    d1[0] = 3
    t1.Fill()
    d1[0] = 2
    t1.Fill()
    d1[0] = 1
    t1.Fill()
    d1[0] = 5
    t1.Fill()

    t1.Write()
    f1.Close()

    b1 = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    b2 = [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]

    with uproot.update(newfile) as fin:
        with pytest.raises(TypeError):
            oldtree = fin["t1"]

        fin.mktree("t2", {"b1": np.int32, "b2": np.float64}, "title")

        for _ in range(5):
            fin["t2"].extend({"b1": b1, "b2": b2})

    f1 = ROOT.TFile(newfile)
    t2root = f1.Get("t2")
    assert t2root.GetEntries() == len(b1) * 5
    assert [x.b1 for x in t2root] == b1 * 5
    assert [x.b2 for x in t2root] == b2 * 5

    t1root = f1.Get("t1")
    assert [x.branch1 for x in t1root] == [5, 4, 3, 2, 1, 5]

    with uproot.open(newfile) as finagin:
        t2uproot = finagin["t2"]
        assert t2uproot.num_entries == len(b1) * 5
        assert t2uproot["b1"].array(library="np").tolist() == b1 * 5
        assert t2uproot["b2"].array(library="np").tolist() == b2 * 5

        assert finagin["t1/branch1"].array(library="np").tolist() == [5, 4, 3, 2, 1, 5]

    f1.Close()


def test_interface(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    branch1 = np.arange(10)
    branch2 = 1.1 * np.arange(10)

    (entries, edges) = np.histogram(branch2)

    with uproot.recreate(newfile) as fout:
        fout["tree"] = {"branch1": branch1, "branch2": branch2}
        fout["tree"].extend({"branch1": branch1, "branch2": branch2})
        fout["hist"] = (entries, edges)

    with uproot.open(newfile) as fin:
        assert fin["tree/branch1"].array(library="np").tolist() == branch1.tolist() * 2
        assert fin["tree/branch2"].array(library="np").tolist() == branch2.tolist() * 2
        assert fin["hist"].to_numpy()[0].tolist() == entries.tolist()
        assert fin["hist"].to_numpy()[1].tolist() == edges.tolist()



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest

import uproot


@pytest.fixture(scope="function", autouse=False)
def reset_classes():
    uproot.model.reset_classes()
    return



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys
from io import StringIO

import numpy
import pytest
import skhep_testdata

import uproot


def test_any_basket():
    interpretation = uproot.interpretation.numerical.AsDtype(">i4")

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample/i4"] as branch:
        assert branch.basket(0).array(interpretation, library="np").tolist() == [
            -15,
            -14,
            -13,
            -12,
            -11,
            -10,
            -9,
        ]
        assert branch.basket(1).array(interpretation, library="np").tolist() == [
            -8,
            -7,
            -6,
            -5,
            -4,
            -3,
            -2,
        ]
        assert branch.basket(2).array(interpretation, library="np").tolist() == [
            -1,
            0,
            1,
            2,
            3,
            4,
            5,
        ]
        assert branch.basket(3).array(interpretation, library="np").tolist() == [
            6,
            7,
            8,
            9,
            10,
            11,
            12,
        ]
        assert branch.basket(4).array(interpretation, library="np").tolist() == [
            13,
            14,
        ]


def test_stitching_arrays():
    interpretation = uproot.interpretation.numerical.AsDtype("i8")
    expectation = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
    basket_arrays = [[0, 1, 2, 3, 4], [5, 6], [], [7, 8, 9], [10], [11, 12, 13, 14]]
    basket_arrays = [numpy.array(x) for x in basket_arrays]
    entry_offsets = numpy.array([0, 5, 7, 7, 10, 11, 15])
    library = uproot.interpretation.library._libraries["np"]

    for start in range(16):
        for stop in range(15, -1, -1):
            actual = interpretation.final_array(
                basket_arrays, start, stop, entry_offsets, library, None, {}
            )
            assert expectation[start:stop] == actual.tolist()


def _names_entries_to_ranges_or_baskets(self, branch_names, entry_start, entry_stop):
    out = []
    for name in branch_names:
        branch = self[name]
        for basket_num, range_or_basket in branch.entries_to_ranges_or_baskets(
            entry_start, entry_stop
        ):
            out.append((branch, basket_num, range_or_basket))
    return out


def test_names_entries_to_ranges_or_baskets():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as sample:
        out = _names_entries_to_ranges_or_baskets(sample, ["i4"], 0, 30)
        assert [x[1] for x in out] == [0, 1, 2, 3, 4]
        assert [x[2] for x in out] == [
            (6992, 7091),
            (16085, 16184),
            (25939, 26038),
            (35042, 35141),
            (40396, 40475),
        ]


def test_ranges_or_baskets_to_arrays():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as sample:
        branch = sample["i4"]

        ranges_or_baskets = _names_entries_to_ranges_or_baskets(sample, ["i4"], 0, 30)
        branchid_interpretation = {
            branch.cache_key: uproot.interpretation.numerical.AsDtype(">i4")
        }
        entry_start, entry_stop = (0, 30)
        decompression_executor = uproot.source.futures.TrivialExecutor()
        interpretation_executor = uproot.source.futures.TrivialExecutor()
        library = uproot.interpretation.library._libraries["np"]

        arrays = {}
        uproot.behaviors.TBranch._ranges_or_baskets_to_arrays(
            sample,
            ranges_or_baskets,
            branchid_interpretation,
            entry_start,
            entry_stop,
            decompression_executor,
            interpretation_executor,
            library,
            arrays,
            False,
            {},
        )
        assert arrays[branch.cache_key].tolist() == [
            -15,
            -14,
            -13,
            -12,
            -11,
            -10,
            -9,
            -8,
            -7,
            -6,
            -5,
            -4,
            -3,
            -2,
            -1,
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
        ]


@pytest.mark.parametrize(
    "file_handler",
    [uproot.source.file.MultithreadedFileSource, uproot.source.file.MemmapSource],
)
def test_branch_array_1(file_handler):
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root"),
        file_handler=file_handler,
    )["sample/i4"] as branch:
        assert branch.array(
            uproot.interpretation.numerical.AsDtype(">i4"), library="np"
        ).tolist() == [
            -15,
            -14,
            -13,
            -12,
            -11,
            -10,
            -9,
            -8,
            -7,
            -6,
            -5,
            -4,
            -3,
            -2,
            -1,
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
        ]


@pytest.mark.parametrize(
    "file_handler",
    [uproot.source.file.MultithreadedFileSource, uproot.source.file.MemmapSource],
)
def test_branch_array_2(file_handler):
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root"),
        file_handler=file_handler,
    )["sample/i4"] as branch:
        assert branch.array(
            uproot.interpretation.numerical.AsDtype(">i4"),
            entry_start=3,
            entry_stop=-5,
            library="np",
        ).tolist() == [
            -12,
            -11,
            -10,
            -9,
            -8,
            -7,
            -6,
            -5,
            -4,
            -3,
            -2,
            -1,
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
        ]


@pytest.mark.parametrize(
    "file_handler",
    [uproot.source.file.MultithreadedFileSource, uproot.source.file.MemmapSource],
)
def test_branch_array_3(file_handler):
    executor = uproot.ThreadPoolExecutor()
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root"),
        file_handler=file_handler,
        interpretation_executor=executor,
        decompression_executor=executor,
    )["sample/i4"] as branch:
        assert branch.array(
            uproot.interpretation.numerical.AsDtype(">i4"),
            entry_start=3,
            entry_stop=-5,
            library="np",
        ).tolist() == [
            -12,
            -11,
            -10,
            -9,
            -8,
            -7,
            -6,
            -5,
            -4,
            -3,
            -2,
            -1,
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
        ]


@pytest.mark.parametrize(
    "file_handler",
    [uproot.source.file.MultithreadedFileSource, uproot.source.file.MemmapSource],
)
def test_branch_array_4(file_handler):
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root"),
        file_handler=file_handler,
    )["sample/i4"] as branch:
        with pytest.raises(ValueError):
            branch.array(uproot.interpretation.numerical.AsDtype(">i8"), library="np")


def test_cache():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root"),
        object_cache=100,
        array_cache="100 MB",
    ) as f:
        assert f.cache_key == "db4be408-93ad-11ea-9027-d201a8c0beef:/"
        assert f["sample"].cache_key == "db4be408-93ad-11ea-9027-d201a8c0beef:/sample;1"
        assert (
            f["sample/i4"].cache_key
            == "db4be408-93ad-11ea-9027-d201a8c0beef:/sample;1:i4(16)"
        )
        i4 = f["sample/i4"]
        assert list(f.file.array_cache) == []
        i4.array(uproot.interpretation.numerical.AsDtype(">i4"), library="np")
        assert list(f.file.array_cache) == [
            "db4be408-93ad-11ea-9027-d201a8c0beef:/sample;1:i4(16):i4:AsDtype(Bi4(),Li4()):0-30:np"
        ]

    with pytest.raises(OSError):
        i4.array(
            uproot.interpretation.numerical.AsDtype(">i4"), entry_start=3, library="np"
        )


def test_pandas():
    pandas = pytest.importorskip("pandas")
    executor = uproot.ThreadPoolExecutor()
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root"),
        interpretation_executor=executor,
        decompression_executor=executor,
    )["sample/i4"] as branch:
        series = branch.array(
            uproot.interpretation.numerical.AsDtype(">i4"),
            entry_start=3,
            entry_stop=-5,
            library="pd",
        )
        assert isinstance(series, pandas.Series)
        assert series.values.tolist() == [
            -12,
            -11,
            -10,
            -9,
            -8,
            -7,
            -6,
            -5,
            -4,
            -3,
            -2,
            -1,
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
        ]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy
import pytest
import skhep_testdata

import uproot


def test_fallback_reading():
    with uproot.open(
        skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root")
    ) as f:
        f["tree:evt/P3/P3.Py"]
        assert f.file._streamers is None



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import copy
import os
import pickle
import sys

import numpy
import pytest
import skhep_testdata

import uproot


def test_detachment():
    with uproot.open(skhep_testdata.data_path("uproot-hepdata-example.root")) as f:
        assert getattr(f["hpx"].file, "file_path", None) is not None
        assert getattr(f["hpx"].file, "source", None) is None

        assert getattr(f["ntuple"].file, "file_path", None) is not None
        assert getattr(f["ntuple"].file, "source", None) is not None

    with uproot.open(
        skhep_testdata.data_path("uproot-small-evnt-tree-nosplit.root")
    ) as f:
        array = f["tree/evt"].array(library="np", entry_stop=1)
        assert getattr(array[0].file, "file_path", None) is not None
        assert getattr(array[0].file, "source", None) is None

        assert isinstance(
            f.file.streamer_named("Event").file, uproot.reading.DetachedFile
        )
        assert (
            str(f.file.streamer_named("Event").file_uuid)
            == "9eebcae8-366b-11e7-ab9d-5e789e86beef"
        )


def test_copy():
    with uproot.open(skhep_testdata.data_path("uproot-hepdata-example.root")) as f:
        original = f["hpx"]
        original_file_path = original.file.file_path

        reconstituted = copy.deepcopy(original)
        reconstituted_file_path = reconstituted.file.file_path

        assert original_file_path == reconstituted_file_path


def test_pickle():
    with uproot.open(skhep_testdata.data_path("uproot-hepdata-example.root")) as f:
        original = f["hpx"]
        original_file_path = original.file.file_path

        reconstituted = pickle.loads(pickle.dumps(original))
        reconstituted_file_path = reconstituted.file.file_path

        assert original_file_path == reconstituted_file_path


def test_pickle_boost():
    boost_histogram = pytest.importorskip("boost_histogram")
    with uproot.open(skhep_testdata.data_path("uproot-hepdata-example.root")) as f:
        original = f["hpx"]
        original_boost = original.to_boost()

        reconstituted = pickle.loads(pickle.dumps(original))
        reconstituted_boost = reconstituted.to_boost()

        pickle.loads(pickle.dumps(original_boost))
        pickle.loads(pickle.dumps(reconstituted_boost))



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import numpy as np
import pytest
import skhep_testdata

import uproot


def test_numpy():
    with uproot.open(skhep_testdata.data_path("uproot-issue-589.root")) as file:
        array = file["events"]["MC_px"].array(library="np", entry_stop=2)
        assert array[0][7] == 30.399463653564453
        assert array[1][11] == 42.04872131347656


pytest.importorskip("awkward")


def test_awkward():
    with uproot.open(skhep_testdata.data_path("uproot-issue-589.root")) as file:
        array = file["events"]["MC_px"].array(library="ak", entry_stop=2)
        assert array[0][7] == 30.399463653564453
        assert array[1][11] == 42.04872131347656



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import pytest

import uproot
import uproot.writing

# THashList should be serialized the same as TList


def test_write_compare_tlist(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    entries = [
        uproot.writing.identify.to_TObjString(s)
        for s in "this is a test string".split()
    ]

    tlist = uproot.writing.identify.to_TList(entries)
    thashlist = uproot.writing.identify.to_THashList(entries)

    with uproot.recreate(filename) as f1:
        f1["tlist"] = tlist
        f1["thashlist"] = thashlist

    with uproot.open(filename) as f2:
        tlist_out = []
        uproot.serialization._serialize_object_any(tlist_out, f2["tlist"], "test")

        thashlist_out = []
        uproot.serialization._serialize_object_any(
            thashlist_out, f2["thashlist"], "test"
        )

        assert (
            tlist_out[0] != thashlist_out[0]
        )  # this section contains the class name which is different
        assert (
            tlist_out[1:] == thashlist_out[1:]
        )  # the remaining bytes should be the same



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines utilities for adding components to the forth reader.
"""

import numpy as np

symbol_dict = {
    np.dtype(">f4"): "f",
    np.dtype(">f8"): "d",
    np.dtype(">i8"): "q",
    np.dtype(">i4"): "i",
    np.dtype(">i2"): "h",
    np.dtype(">u1"): "B",
    np.dtype(">u2"): "H",
    np.dtype(">u4"): "I",
    np.dtype(">u8"): "Q",
    np.dtype("bool"): "?",
}


class ForthGenerator:
    """
    This class is passed through the Forth code generation, collecting Forth snippets and concatenating them at the end.
    """

    def __init__(self, aform=None, count_obj=0, var_set=False):
        self.dummy_form = False
        self.top_dummy = None
        self.dummy_aform = None
        self.aform = aform
        self.top_form = None
        self.prev_form = None
        self.awkward_model = {"name": "TOP", "content": {}}
        self.top_node = self.awkward_model
        self.ref_list = []
        self.forth_code = {}
        self.forth_keys = {}
        self.final_code = []
        self.final_header = []
        self.final_init = []
        self.form_keys = []
        self.var_set = var_set
        self.count_obj = count_obj

    def traverse_aform(self):
        self.aform = self.aform.content

    def replace_form_and_model(self, form, model):
        temp_node = self.awkward_model
        temp_prev_form = self.prev_form
        temp_node_top = self.top_node
        self.awkward_model = model
        self.top_node = self.awkward_model
        temp_form = self.aform
        temp_form_top = self.top_form
        self.top_form = None
        self.aform = form
        return temp_node, temp_node_top, temp_form, temp_form_top, temp_prev_form

    def get_code_recursive(self, node):
        pre, post, init, header = self.tree_walk(node)
        return pre, post, init, header

    def tree_walk(self, node):
        if "content" in node.keys():
            if node["content"] is None:
                return (
                    "".join(node["pre_code"]),
                    "".join(node["post_code"]),
                    node["init_code"],
                    node["header_code"],
                )
            else:
                pre, post, init, header = self.tree_walk(node["content"])
                pre2 = "".join(node["pre_code"])
                pre2 = pre2 + pre
                post2 = "".join(node["post_code"])
                post2 = post2 + post
                init = node["init_code"] + init
                header = node["header_code"] + header
                return pre2 + post2, "", init, header
        elif self.var_set:
            return "", "", "", ""

    def should_add_form(self):
        if "content" in self.awkward_model.keys():
            if self.awkward_model["content"] is None:
                return False
            else:
                if len(self.awkward_model["content"].keys()) == 0:
                    return True
                elif self.awkward_model["content"]["name"] == "dynamic":
                    return True
                else:
                    return False

    def get_temp_form_top(self):
        return self.top_dummy

    def add_form(self, aform, conlen=0, traverse=1):
        if self.aform is None:
            self.aform = aform
            self.top_form = self.aform
            if traverse == 2:
                self.aform = self.aform["content"]
        else:
            if "content" in self.aform.keys():
                if self.aform["content"] == "NULL":
                    self.aform["content"] = aform
                    self.prev_form = self.aform
                    if traverse == 2:
                        self.aform = self.aform["content"]["content"]
                        self.prev_form = self.prev_form["content"]
                    else:
                        self.aform = self.aform["content"]
                else:
                    raise AssertionError("Form representation is corrupted.")
            elif "contents" in self.aform.keys():
                if self.aform["class"] == "RecordArray":
                    if self.prev_form is not None:
                        self.prev_form["content"] = aform
                        self.aform = aform
                    else:
                        self.top_form = aform
                        self.aform = aform
                elif len(self.aform["contents"]) == conlen:
                    pass
                else:
                    return False

    def set_dummy_none(self, temp_top, temp_form, temp_flag):
        self.top_dummy = temp_top
        self.dummy_aform = temp_form
        self.dummy_form = temp_flag

    def get_keys(self, num_keys):
        if num_keys == 1:
            key = self.count_obj
            self.count_obj += 1
            return key
        elif num_keys > 1:
            out = []
            for _i in range(num_keys):
                out.append(self.count_obj)
                self.count_obj += 1
            return out
        else:
            raise AssertionError("Number of keys cannot be less than 1")

    def add_form_key(self, key):
        self.form_keys.append(key)

    def go_to(self, aform):
        self.awkward_model = aform

    def become(self, aform):
        self.awkward_model = aform

    def check_model(self):
        return bool(self.awkward_model)

    def get_current_node(self):
        self.ref_list.append(self.top_node)
        return len(self.ref_list) - 1

    def get_ref(self, index):
        return self.ref_list[index]

    def enable_adding(self):
        if "content" in self.awkward_model.keys():
            if self.awkward_model["content"] is None:
                self.awkward_model["content"] = {}

    def add_node_whole(self, new_node, ref_latest):
        if "content" in self.awkward_model.keys():
            self.awkward_model["content"] = new_node
        else:
            self.awkward_model["contents"].append(new_node)
        self.awkward_model = ref_latest

    def add_node(self, name, code_attrs, dtype, num_child, content):
        if isinstance(self.awkward_model, dict):
            if (
                not bool(self.awkward_model["content"])
                and self.awkward_model["content"] is not None
            ):
                temp_obj = {
                    "name": name,
                    "type": dtype,
                    "pre_code": code_attrs["precode"],
                    "post_code": code_attrs["postcode"],
                    "init_code": code_attrs["initcode"],
                    "header_code": code_attrs["headercode"],
                    "num_child": num_child,
                    "content": content,
                }
                self.awkward_model["content"] = temp_obj
                self.awkward_model = self.awkward_model["content"]
                return temp_obj
            else:
                if self.awkward_model["content"] is not None:
                    if self.awkward_model["content"]["name"] == "dynamic":
                        temp_obj = {
                            "name": name,
                            "type": dtype,
                            "pre_code": code_attrs["precode"],
                            "post_code": code_attrs["postcode"],
                            "init_code": code_attrs["initcode"],
                            "header_code": code_attrs["headercode"],
                            "num_child": num_child,
                            "content": content,
                        }
                        self.awkward_model["content"] = temp_obj
                        self.awkward_model = self.awkward_model["content"]
                        return temp_obj
                    else:
                        temp_node = self.awkward_model
                        self.awkward_model = self.awkward_model["content"]
                        return temp_node
                else:
                    temp_node = self.awkward_model
                    self.awkward_model = self.awkward_model["content"]
                    return temp_node
        if isinstance(self.awkward_model, list):
            for elem in self.awkward_model:
                if name in elem.values():
                    return elem
            self.awkward_model.append(
                {
                    "name": name,
                    "type": dtype,
                    "pre_code": code_attrs["precode"],
                    "post_code": code_attrs["postcode"],
                    "init_code": code_attrs["initcode"],
                    "header_code": code_attrs["headercode"],
                    "num_child": num_child,
                    "content": content,
                }
            )
            temp_node = self.awkward_model[-1]
            self.awkward_model = self.awkward_model[-1]["content"]
            return temp_node

    def register_pre(self, ref):
        key = str(ref) + "pre"
        self.forth_sequence.append(key)

    def add_meta(self, ref, form_key, header, init):
        if self.forth_code[id(ref)] is None:
            self.forth_code[id(ref)] = {}
        for elem in form_key:
            self.form_keys.append(elem)
        self.forth_code[id(ref)]["forth_header"] = header
        self.forth_code[id(ref)]["forth_init"] = init

    def register_post(self, ref):
        key = str(ref) + "post"
        self.forth_sequence.append(key)

    def add_forth_code(self, ref, forth_exec_pre, forth_exec_post):
        if self.forth_code[ref] is None:
            self.forth_code[ref] = {}
        self.forth_code[ref][str(ref) + "pre"] = forth_exec_pre
        self.forth_code[ref][str(ref) + "post"] = forth_exec_post

    def add_to_final(self, code):
        if not isinstance(code, list):
            raise AssertionError("Something went wrong with Forth code generation.")
        self.final_code.extend(code)

    def add_to_header(self, code):
        self.final_header.append(code)

    def add_to_init(self, code):
        self.final_init.append(code)


def forth_stash(context):
    """
    Returns a ForthLevelStash object if ForthGeneration is to be done, else None.
    """
    if hasattr(context.get("forth"), "gen"):
        return ForthLevelStash(context["forth"].gen)
    else:
        return None


class ForthLevelStash:
    """
    Helper class to stash code at one level of Forth code generation. Keeps the code generation clean and maintains order for the code snippets.
    """

    def __init__(self, context):
        self._pre_code = []
        self._post_code = []
        self._header = ""
        self._init = ""
        self._form_key = []
        self._gen_obj = context

    def is_forth(self):
        return self.forth_present

    def get_gen_obj(self):
        return self._gen_obj

    def add_to_pre(self, code):
        self._pre_code.append(code)

    def add_to_post(self, code):
        self._post_code.append(code)

    def get_attrs(self):
        return {
            "precode": self._pre_code,
            "postcode": self._post_code,
            "initcode": self._init,
            "headercode": self._header,
        }

    def add_to_header(self, code):
        self._header += code

    def add_to_init(self, code):
        self._init += code


def convert_dtype(format):
    """Takes datatype codes from classses and returns the full datatype name.

    Args:
        format (string): The datatype in the dynamic class

    Returns:
        string: The datatype in words.
    """
    if format == "d":
        return "float64"
    elif format == "f":
        return "float32"
    elif format == "q":
        return "int64"
    elif format == "i":
        return "int32"
    elif format == "I":
        return "uint32"
    elif format == "?":
        return "bool"
    elif format == "h":
        return "int16"
    elif format == "H":
        return "uint16"
    elif format == "Q":
        return "uint64"
    elif format == "B":
        return "uint8"



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines a versionless model for ``TObjString``.
"""


import uproot
import uproot.serialization


class Model_TObjString(uproot.model.Model, str):
    """
    A versionless :doc:`uproot.model.Model` for ``TObjString``.

    This is also a Python ``str`` (string).
    """

    class_rawstreamers = (
        (
            None,
            b"@\x00\x01X\xff\xff\xff\xffTStreamerInfo\x00@\x00\x01B\x00\t@\x00\x00\x18\x00\x01\x00\x01\x00\x00\x00\x00\x03\x01\x00\x00\nTObjString\x00\x9c\x8eH\x00\x00\x00\x00\x01@\x00\x01\x18\xff\xff\xff\xffTObjArray\x00@\x00\x01\x06\x00\x03\x00\x01\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00@\x00\x00u\xff\xff\xff\xffTStreamerBase\x00@\x00\x00_\x00\x03@\x00\x00U\x00\x04@\x00\x00&\x00\x01\x00\x01\x00\x00\x00\x00\x03\x00\x00\x00\x07TObject\x11Basic ROOT object\x00\x00\x00B\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x90\x1b\xc0-\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04BASE\x00\x00\x00\x01@\x00\x00t\xff\xff\xff\xffTStreamerString\x00@\x00\x00\\\x00\x02@\x00\x00V\x00\x04@\x00\x00$\x00\x01\x00\x01\x00\x00\x00\x00\x03\x00\x00\x00\x07fString\x0fwrapped TString\x00\x00\x00A\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07TString\x00",
            "TObjString",
            1,
        ),
    )
    writable = True

    def read_members(self, chunk, cursor, context, file):
        context["cancel_forth"] = True
        if self.is_memberwise:
            raise NotImplementedError(
                """memberwise serialization of {}
in file {}""".format(
                    type(self).__name__, self.file.file_path
                )
            )
        self._bases.append(
            uproot.models.TObject.Model_TObject.read(
                chunk,
                cursor,
                context,
                file,
                self._file,
                self._parent,
                concrete=self.concrete,
            )
        )
        self._data = cursor.string(chunk, context)

    def postprocess(self, chunk, cursor, context, file):
        out = Model_TObjString(self._data)
        out._cursor = self._cursor
        out._file = self._file
        out._parent = self._parent
        out._members = self._members
        out._bases = self._bases
        out._num_bytes = self._num_bytes
        out._instance_version = self._instance_version
        return out

    @property
    def fTitle(self):
        return "Collectable string class"

    writable = True

    def tojson(self):
        out = self._bases[0].tojson()  # TObject
        out["_typename"] = self.classname
        out["fString"] = str(self)
        return out

    def _serialize(self, out, header, name, tobject_flags):
        where = len(out)
        for x in self._bases:
            x._serialize(out, True, name, tobject_flags | uproot.const.kNotDeleted)
        out.append(uproot.serialization.string(str(self)))
        if header:
            num_bytes = sum(len(x) for x in out[where:])
            version = 1
            out.insert(where, uproot.serialization.numbytes_version(num_bytes, version))

    def __repr__(self):
        if self.class_version is None:
            version = ""
        else:
            version = f" (version {self.class_version})"
        return "<{}{} {} at 0x{:012x}>".format(
            self.classname,
            version,
            str.__repr__(self),
            id(self),
        )

    @classmethod
    def awkward_form(cls, file, context):
        awkward = uproot.extras.awkward()
        return awkward.forms.ListOffsetForm(
            context["index_format"],
            awkward.forms.NumpyForm("uint8", parameters={"__array__": "char"}),
            parameters={"__array__": "string"},
        )


uproot.classes["TObjString"] = Model_TObjString



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest
import skhep_testdata

import uproot


def test_keys():
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree"
    ] as t:
        assert t.keys(filter_name="P3.Px") == ["evt/P3/P3.Px"]
        assert t.keys(filter_name="/P3.Px") == []
        assert t.keys(filter_name="P3/P3.Px") == []
        assert t.keys(filter_name="evt/P3/P3.Px") == ["evt/P3/P3.Px"]
        assert t.keys(filter_name="/evt/P3/P3.Px") == ["evt/P3/P3.Px"]
        assert t["evt"].keys(filter_name="P3.Px") == ["P3/P3.Px"]
        assert t["evt"].keys(filter_name="/P3.Px") == []
        assert t["evt"].keys(filter_name="P3/P3.Px") == ["P3/P3.Px"]
        assert t["evt"].keys(filter_name="/P3/P3.Px") == ["P3/P3.Px"]
        assert t["evt"].keys(filter_name="evt/P3/P3.Px") == []
        assert t["evt/P3"].keys(filter_name="P3.Px") == ["P3.Px"]
        assert t["evt/P3"].keys(filter_name="/P3.Px") == ["P3.Px"]
        assert t["evt/P3"].keys(filter_name="P3/P3.Px") == []


def test_numpy():
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree"
    ] as t:
        assert list(t.arrays(filter_name="P3.Px", library="np").keys()) == ["P3.Px"]
        assert list(t.arrays(filter_name="/P3.Px", library="np").keys()) == []
        assert list(t.arrays(filter_name="P3/P3.Px", library="np").keys()) == []
        assert list(t.arrays(filter_name="evt/P3/P3.Px", library="np").keys()) == [
            "P3.Px"
        ]
        assert list(t.arrays(filter_name="/evt/P3/P3.Px", library="np").keys()) == [
            "P3.Px"
        ]
        assert list(t["evt"].arrays(filter_name="P3.Px", library="np").keys()) == [
            "P3.Px"
        ]
        assert list(t["evt"].arrays(filter_name="/P3.Px", library="np").keys()) == []
        assert list(t["evt"].arrays(filter_name="P3/P3.Px", library="np").keys()) == [
            "P3.Px"
        ]
        assert list(t["evt"].arrays(filter_name="/P3/P3.Px", library="np").keys()) == [
            "P3.Px"
        ]
        assert (
            list(t["evt"].arrays(filter_name="evt/P3/P3.Px", library="np").keys()) == []
        )
        assert list(t["evt/P3"].arrays(filter_name="P3.Px", library="np").keys()) == [
            "P3.Px"
        ]
        assert list(t["evt/P3"].arrays(filter_name="/P3.Px", library="np").keys()) == [
            "P3.Px"
        ]
        assert (
            list(t["evt/P3"].arrays(filter_name="P3/P3.Px", library="np").keys()) == []
        )

        assert list(t.arrays("P3.Px", library="np").keys()) == ["P3.Px"]
        with pytest.raises(Exception):
            t.arrays("/P3.Px", library="np")
        assert list(t.arrays("evt/P3/P3.Px", library="np").keys()) == ["evt/P3/P3.Px"]
        assert list(t.arrays("/evt/P3/P3.Px", library="np").keys()) == ["/evt/P3/P3.Px"]
        assert list(t["evt"].arrays("P3.Px", library="np").keys()) == ["P3.Px"]
        with pytest.raises(Exception):
            t["evt"].arrays("/P3.Px", library="np")
        assert list(t["evt"].arrays("P3/P3.Px", library="np").keys()) == ["P3/P3.Px"]
        assert list(t["evt/P3"].arrays("P3.Px", library="np").keys()) == ["P3.Px"]
        assert list(t["evt/P3"].arrays("/P3.Px", library="np").keys()) == ["/P3.Px"]


def test_awkward():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree"
    ] as t:
        assert t.arrays(filter_name="P3.Px", library="ak").fields == ["P3.Px"]
        assert t.arrays(filter_name="/P3.Px", library="ak").fields == []
        assert t.arrays(filter_name="P3/P3.Px", library="ak").fields == []
        assert t.arrays(filter_name="evt/P3/P3.Px", library="ak").fields == ["P3.Px"]
        assert t.arrays(filter_name="/evt/P3/P3.Px", library="ak").fields == ["P3.Px"]
        assert t["evt"].arrays(filter_name="P3.Px", library="ak").fields == ["P3.Px"]
        assert t["evt"].arrays(filter_name="/P3.Px", library="ak").fields == []
        assert t["evt"].arrays(filter_name="P3/P3.Px", library="ak").fields == ["P3.Px"]
        assert t["evt"].arrays(filter_name="/P3/P3.Px", library="ak").fields == [
            "P3.Px"
        ]
        assert t["evt"].arrays(filter_name="evt/P3/P3.Px", library="ak").fields == []
        assert t["evt/P3"].arrays(filter_name="P3.Px", library="ak").fields == ["P3.Px"]
        assert t["evt/P3"].arrays(filter_name="/P3.Px", library="ak").fields == [
            "P3.Px"
        ]
        assert t["evt/P3"].arrays(filter_name="P3/P3.Px", library="ak").fields == []

        assert t.arrays("P3.Px", library="ak").fields == ["P3.Px"]
        with pytest.raises(Exception):
            t.arrays("/P3.Px", library="ak")
        assert t.arrays("evt/P3/P3.Px", library="ak").fields == ["evt/P3/P3.Px"]
        assert t.arrays("/evt/P3/P3.Px", library="ak").fields == ["/evt/P3/P3.Px"]
        assert t["evt"].arrays("P3.Px", library="ak").fields == ["P3.Px"]
        with pytest.raises(Exception):
            t["evt"].arrays("/P3.Px", library="ak")
        assert t["evt"].arrays("P3/P3.Px", library="ak").fields == ["P3/P3.Px"]
        assert t["evt/P3"].arrays("P3.Px", library="ak").fields == ["P3.Px"]
        assert t["evt/P3"].arrays("/P3.Px", library="ak").fields == ["/P3.Px"]


def test_pandas():
    pandas = pytest.importorskip("pandas")
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree"
    ] as t:
        assert t.arrays(filter_name="P3.Px", library="pd").columns.tolist() == ["P3.Px"]
        assert t.arrays(filter_name="/P3.Px", library="pd").columns.tolist() == []
        assert t.arrays(filter_name="P3/P3.Px", library="pd").columns.tolist() == []
        assert t.arrays(filter_name="evt/P3/P3.Px", library="pd").columns.tolist() == [
            "P3.Px"
        ]
        assert t.arrays(filter_name="/evt/P3/P3.Px", library="pd").columns.tolist() == [
            "P3.Px"
        ]
        assert t["evt"].arrays(filter_name="P3.Px", library="pd").columns.tolist() == [
            "P3.Px"
        ]
        assert (
            t["evt"].arrays(filter_name="/P3.Px", library="pd").columns.tolist() == []
        )
        assert t["evt"].arrays(
            filter_name="P3/P3.Px", library="pd"
        ).columns.tolist() == ["P3.Px"]
        assert t["evt"].arrays(
            filter_name="/P3/P3.Px", library="pd"
        ).columns.tolist() == ["P3.Px"]
        assert (
            t["evt"].arrays(filter_name="evt/P3/P3.Px", library="pd").columns.tolist()
            == []
        )
        assert t["evt/P3"].arrays(
            filter_name="P3.Px", library="pd"
        ).columns.tolist() == ["P3.Px"]
        assert t["evt/P3"].arrays(
            filter_name="/P3.Px", library="pd"
        ).columns.tolist() == ["P3.Px"]
        assert (
            t["evt/P3"].arrays(filter_name="P3/P3.Px", library="pd").columns.tolist()
            == []
        )

        assert t.arrays("P3.Px", library="pd").columns.tolist() == ["P3.Px"]
        with pytest.raises(Exception):
            t.arrays("/P3.Px", library="pd")
        assert t.arrays("evt/P3/P3.Px", library="pd").columns.tolist() == [
            "evt/P3/P3.Px"
        ]
        assert t.arrays("/evt/P3/P3.Px", library="pd").columns.tolist() == [
            "/evt/P3/P3.Px"
        ]
        assert t["evt"].arrays("P3.Px", library="pd").columns.tolist() == ["P3.Px"]
        with pytest.raises(Exception):
            t["evt"].arrays("/P3.Px", library="pd")
        assert t["evt"].arrays("P3/P3.Px", library="pd").columns.tolist() == [
            "P3/P3.Px"
        ]
        assert t["evt/P3"].arrays("P3.Px", library="pd").columns.tolist() == ["P3.Px"]
        assert t["evt/P3"].arrays("/P3.Px", library="pd").columns.tolist() == ["/P3.Px"]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os
import queue
import sys
from io import StringIO

import numpy
import pytest

import uproot


def test_file(tmpdir):
    filename = os.path.join(str(tmpdir), "tmp.raw")
    with open(filename, "wb") as tmp:
        tmp.write(b"******    ...+++++++!!!!!@@@@@")

    notifications = queue.Queue()
    with uproot.source.file.MultithreadedFileSource(filename, num_workers=1) as source:
        chunks = source.chunks(
            [(0, 6), (6, 10), (10, 13), (13, 20), (20, 25), (25, 30)],
            notifications=notifications,
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))


def test_file_workers(tmpdir):
    filename = os.path.join(str(tmpdir), "tmp.raw")
    with open(filename, "wb") as tmp:
        tmp.write(b"******    ...+++++++!!!!!@@@@@")

    notifications = queue.Queue()
    with uproot.source.file.MultithreadedFileSource(filename, num_workers=5) as source:
        chunks = source.chunks(
            [(0, 6), (6, 10), (10, 13), (13, 20), (20, 25), (25, 30)],
            notifications=notifications,
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))


def test_memmap(tmpdir):
    filename = os.path.join(str(tmpdir), "tmp.raw")
    with open(filename, "wb") as tmp:
        tmp.write(b"******    ...+++++++!!!!!@@@@@")

    notifications = queue.Queue()
    with uproot.source.file.MemmapSource(filename, num_fallback_workers=1) as source:
        chunks = source.chunks(
            [(0, 6), (6, 10), (10, 13), (13, 20), (20, 25), (25, 30)],
            notifications=notifications,
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))


@pytest.mark.skip(reason="RECHECK: example.com is flaky, too")
@pytest.mark.network
def test_http_multipart():
    notifications = queue.Queue()
    with uproot.source.http.HTTPSource(
        "https://example.com", timeout=10, num_fallback_workers=1
    ) as source:
        chunks = source.chunks(
            [(0, 100), (50, 55), (200, 400)], notifications=notifications
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))


@pytest.mark.skip(reason="RECHECK: example.com is flaky, too")
@pytest.mark.network
def test_http():
    notifications = queue.Queue()
    with uproot.source.http.MultithreadedHTTPSource(
        "https://example.com", timeout=10, num_workers=1
    ) as source:
        chunks = source.chunks(
            [(0, 100), (50, 55), (200, 400)], notifications=notifications
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))


@pytest.mark.skip(reason="RECHECK: example.com is flaky, too")
@pytest.mark.network
def test_http_workers():
    notifications = queue.Queue()
    with uproot.source.http.MultithreadedHTTPSource(
        "https://example.com", timeout=10, num_workers=2
    ) as source:
        chunks = source.chunks(
            [(0, 100), (50, 55), (200, 400)], notifications=notifications
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))


@pytest.mark.network
def test_http_fallback():
    notifications = queue.Queue()
    with uproot.source.http.HTTPSource(
        "https://scikit-hep.org/uproot3/examples/Zmumu.root",
        timeout=10,
        num_fallback_workers=1,
    ) as source:
        chunks = source.chunks(
            [(0, 100), (50, 55), (200, 400)], notifications=notifications
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))


@pytest.mark.network
def test_http_fallback_workers():
    notifications = queue.Queue()
    with uproot.source.http.HTTPSource(
        "https://scikit-hep.org/uproot3/examples/Zmumu.root",
        timeout=10,
        num_fallback_workers=5,
    ) as source:
        chunks = source.chunks(
            [(0, 100), (50, 55), (200, 400)], notifications=notifications
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd():
    pytest.importorskip("XRootD")
    notifications = queue.Queue()
    with uproot.source.xrootd.MultithreadedXRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        num_workers=1,
        timeout=10,
    ) as source:
        chunks = source.chunks(
            [(0, 100), (50, 55), (200, 400)], notifications=notifications
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_workers():
    pytest.importorskip("XRootD")
    notifications = queue.Queue()
    with uproot.source.xrootd.MultithreadedXRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        num_workers=5,
        timeout=10,
    ) as source:
        chunks = source.chunks(
            [(0, 100), (50, 55), (200, 400)], notifications=notifications
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_vectorread():
    pytest.importorskip("XRootD")
    notifications = queue.Queue()
    with uproot.source.xrootd.XRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        timeout=10,
        max_num_elements=None,
        num_workers=1,
    ) as source:
        chunks = source.chunks(
            [(0, 100), (50, 55), (200, 400)], notifications=notifications
        )
        expected = {(chunk.start, chunk.stop): chunk for chunk in chunks}
        while len(expected) > 0:
            chunk = notifications.get()
            expected.pop((chunk.start, chunk.stop))



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy
import pytest
import skhep_testdata

import uproot


def test_inplace_tbranch_array():
    with uproot.open(skhep_testdata.data_path("uproot-Zmumu.root"))["events"] as events:
        N = events.num_entries

        m_array = numpy.zeros(N, dtype=numpy.float64)
        m_i = events["M"].interpretation.inplace(m_array)
        m_res = events["M"].array(interpretation=m_i, library="np")

        assert m_res.base is m_array
        assert m_res[:5].tolist() == [
            82.4626915551,
            83.6262040052,
            83.3084646667,
            82.1493728809,
            90.4691230355,
        ]

        run_array = numpy.zeros(N, dtype=numpy.int64)
        run_i = events["Run"].interpretation.inplace(run_array)
        run_res = events["Run"].array(interpretation=run_i, library="np")

        assert run_res.base is run_array
        assert run_res[:5].tolist() == [148031, 148031, 148031, 148031, 148031]


def test_inplace_ttree_arrays():
    with uproot.open(skhep_testdata.data_path("uproot-Zmumu.root"))["events"] as events:
        N = events.num_entries

        m_array = numpy.zeros(N, dtype=numpy.float64)
        m_i = events["M"].interpretation.inplace(m_array)

        run_array = numpy.zeros(N, dtype=numpy.int64)
        run_i = events["Run"].interpretation.inplace(run_array)

        events.arrays(dict(M=m_i, Run=run_i), library="np")

        assert m_array[:5].tolist() == [
            82.4626915551,
            83.6262040052,
            83.3084646667,
            82.1493728809,
            90.4691230355,
        ]

        assert run_array[:5].tolist() == [148031, 148031, 148031, 148031, 148031]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys

import numpy
import pytest
import skhep_testdata

import uproot
from uproot.containers import AsMap, AsSet, AsString, AsVector
from uproot.interpretation.identify import parse_typename


def test_parse_typename():
    assert parse_typename("TTree") is uproot.classes["TTree"]
    assert parse_typename("string") == AsString(False)
    assert parse_typename("std::string") == AsString(False)
    assert parse_typename("std :: string") == AsString(False)
    assert parse_typename("char*") == AsString(False, length_bytes="4")
    assert parse_typename("char *") == AsString(False, length_bytes="4")
    assert parse_typename("TString") == AsString(False)
    assert parse_typename("vector<TTree>") == AsVector(True, uproot.classes["TTree"])
    assert parse_typename("vector<int>") == AsVector(True, ">i4")
    assert parse_typename("vector<bool>") == AsVector(True, "?")
    assert parse_typename("vector<string>") == AsVector(True, AsString(False))
    assert parse_typename("vector  <   string   >") == AsVector(True, AsString(False))
    assert parse_typename("std::vector<std::string>") == AsVector(True, AsString(False))
    assert parse_typename("vector<vector<int>>") == AsVector(
        True, AsVector(False, ">i4")
    )
    assert parse_typename("vector<vector<string>>") == AsVector(
        True, AsVector(False, AsString(False))
    )
    assert parse_typename("vector<vector<char*>>") == AsVector(
        True, AsVector(False, AsString(False, length_bytes="4"))
    )
    assert parse_typename("set<unsigned short>") == AsSet(True, ">u2")
    assert parse_typename("std::set<unsigned short>") == AsSet(True, ">u2")
    assert parse_typename("set<string>") == AsSet(True, AsString(False))
    assert parse_typename("set<vector<string>>") == AsSet(
        True, AsVector(False, AsString(False))
    )
    assert parse_typename("set<vector<string> >") == AsSet(
        True, AsVector(False, AsString(False))
    )
    assert parse_typename("map<int, double>") == AsMap(True, ">i4", ">f8")
    assert parse_typename("map<string, double>") == AsMap(True, AsString(True), ">f8")
    assert parse_typename("map<int, string>") == AsMap(True, ">i4", AsString(True))
    assert parse_typename("map<string, string>") == AsMap(
        True, AsString(True), AsString(True)
    )
    assert parse_typename("map<string,string>") == AsMap(
        True, AsString(True), AsString(True)
    )
    assert parse_typename("map<   string,string   >") == AsMap(
        True, AsString(True), AsString(True)
    )
    assert parse_typename("map<string,vector<int>>") == AsMap(
        True, AsString(True), AsVector(True, ">i4")
    )
    assert parse_typename("map<vector<int>, string>") == AsMap(
        True, AsVector(True, ">i4"), AsString(True)
    )
    assert parse_typename("map<vector<int>, set<float>>") == AsMap(
        True, AsVector(True, ">i4"), AsSet(True, ">f4")
    )
    assert parse_typename("map<vector<int>, set<set<float>>>") == AsMap(
        True, AsVector(True, ">i4"), AsSet(True, AsSet(False, ">f4"))
    )

    with pytest.raises(ValueError):
        parse_typename("string  <")

    with pytest.raises(ValueError):
        parse_typename("vector  <")

    with pytest.raises(ValueError):
        parse_typename("map<string<int>>")

    with pytest.raises(ValueError):
        parse_typename("map<string, int>>")


def test_strings1():
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree"
    ] as tree:
        result = tree["Beg"].array(library="np")
        assert result.tolist() == [f"beg-{i:03d}" for i in range(100)]

        result = tree["End"].array(library="np")
        assert result.tolist() == [f"end-{i:03d}" for i in range(100)]


def test_map_string_string_in_object():
    with uproot.open(skhep_testdata.data_path("uproot-issue431.root")) as f:
        head = f["Head"]
        assert head.member("map<string,string>") == {
            "DAQ": "394",
            "PDF": "4      58",
            "XSecFile": "",
            "can": "0 1027 888.4",
            "can_user": "0.00 1027.00  888.40",
            "coord_origin": "0 0 0",
            "cut_in": "0 0 0 0",
            "cut_nu": "100 1e+08 -1 1",
            "cut_primary": "0 0 0 0",
            "cut_seamuon": "0 0 0 0",
            "decay": "doesnt happen",
            "detector": "NOT",
            "drawing": "Volume",
            "end_event": "",
            "genhencut": "2000 0",
            "genvol": "0 1027 888.4 2.649e+09 100000",
            "kcut": "2",
            "livetime": "0 0",
            "model": "1       2       0       1      12",
            "muon_desc_file": "",
            "ngen": "0.1000E+06",
            "norma": "0 0",
            "nuflux": "0       3       0 0.500E+00 0.000E+00 0.100E+01 0.300E+01",
            "physics": "GENHEN 7.2-220514 181116 1138",
            "seed": "GENHEN 3  305765867         0         0",
            "simul": "JSirene 11012 11/17/18 07",
            "sourcemode": "diffuse",
            "spectrum": "-1.4",
            "start_run": "1",
            "target": "isoscalar",
            "usedetfile": "false",
            "xlat_user": "0.63297",
            "xparam": "OFF",
            "zed_user": "0.00 3450.00",
        }


def test_map_long_int_in_object():
    with uproot.open(skhep_testdata.data_path("uproot-issue283.root")) as f:
        map_long_int = f["config/detector"].member("ChannelIDMap")
        assert (map_long_int.keys().min(), map_long_int.keys().max()) == (
            46612627560,
            281410180683757,
        )
        assert (map_long_int.values().min(), map_long_int.values().max()) == (0, 5159)


def test_top_level_vectors():
    with uproot.open(skhep_testdata.data_path("uproot-issue38a.root"))[
        "ntupler/tree"
    ] as tree:
        assert [x.tolist() for x in tree["v_int16"].array(library="np")] == [[1, 2, 3]]
        assert [x.tolist() for x in tree["v_int16"].array(library="np")] == [[1, 2, 3]]
        assert [x.tolist() for x in tree["v_int32"].array(library="np")] == [[1, 2, 3]]
        assert [x.tolist() for x in tree["v_int64"].array(library="np")] == [[1, 2, 3]]
        assert [x.tolist() for x in tree["v_uint16"].array(library="np")] == [[1, 2, 3]]
        assert [x.tolist() for x in tree["v_uint32"].array(library="np")] == [[1, 2, 3]]
        assert [x.tolist() for x in tree["v_uint64"].array(library="np")] == [[1, 2, 3]]
        assert [x.tolist() for x in tree["v_bool"].array(library="np")] == [
            [False, True]
        ]
        assert [x.tolist() for x in tree["v_float"].array(library="np")] == [
            [999.0, -999.0]
        ]
        assert [x.tolist() for x in tree["v_double"].array(library="np")] == [
            [999.0, -999.0]
        ]


def test_strings1():
    with uproot.open(skhep_testdata.data_path("uproot-issue31.root"))[
        "T/name"
    ] as branch:
        result = branch.array(library="np")
        assert result.tolist() == ["one", "two", "three", "four", "five"]


def test_strings2():
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree/Str"
    ] as branch:
        result = branch.array(library="np")
        assert result.tolist() == [f"evt-{i:03d}" for i in range(100)]


def test_strings3():
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree/StdStr"
    ] as branch:
        result = branch.array(library="np")
        assert result.tolist() == [f"std-{i:03d}" for i in range(100)]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines a versionless model of ``TString``.
"""


import uproot
import uproot._awkward_forth


class Model_TString(uproot.model.Model, str):
    """
    A versionless :doc:`uproot.model.Model` for ``TString``.

    This is also a Python ``str`` (string).
    """

    def read_numbytes_version(self, chunk, cursor, context):
        pass

    def read_members(self, chunk, cursor, context, file):
        forth_stash = uproot._awkward_forth.forth_stash(context)
        if forth_stash is not None:
            forth_obj = forth_stash.get_gen_obj()
            keys = forth_obj.get_keys(2)
            offsets_num = keys[0]
            data_num = keys[1]
        if self.is_memberwise:
            raise NotImplementedError(
                """memberwise serialization of {}
in file {}""".format(
                    type(self).__name__, self.file.file_path
                )
            )
        if forth_stash is not None:
            forth_stash.add_to_pre(
                f" stream !B-> stack dup 255 = if drop stream !I-> stack then dup node{offsets_num}-offsets +<- stack stream #!B-> node{data_num}-data\n"
            )
            if forth_obj.should_add_form():
                temp_aform = {
                    "class": "ListOffsetArray",
                    "offsets": "i64",
                    "content": {
                        "class": "NumpyArray",
                        "primitive": "uint8",
                        "inner_shape": [],
                        "parameters": {"__array__": "char"},
                        "form_key": f"node{data_num}",
                    },
                    "parameters": {"__array__": "string"},
                    "form_key": f"node{offsets_num}",
                }
                forth_obj.add_form(temp_aform)

                form_keys = [
                    f"node{data_num}-data",
                    f"node{offsets_num}-offsets",
                ]
                for elem in form_keys:
                    forth_obj.add_form_key(elem)
            forth_stash.add_to_header(
                f"output node{offsets_num}-offsets int64\noutput node{data_num}-data uint8\n"
            )
            forth_stash.add_to_init(f"0 node{offsets_num}-offsets <- stack\n")
            temp_form = forth_obj.add_node(
                f"node{offsets_num}",
                forth_stash.get_attrs(),
                "i64",
                0,
                None,
            )
            forth_obj.go_to(temp_form)
        self._data = cursor.string(chunk, context)

    def postprocess(self, chunk, cursor, context, file):
        out = Model_TString(self._data)
        out._cursor = self._cursor
        out._file = self._file
        out._parent = self._parent
        out._members = self._members
        out._bases = self._bases
        out._num_bytes = self._num_bytes
        out._instance_version = self._instance_version
        return out

    def __repr__(self):
        if self.class_version is None:
            version = ""
        else:
            version = f" (version {self.class_version})"
        return "<{}{} {} at 0x{:012x}>".format(
            self.classname, version, str.__repr__(self), id(self)
        )

    def tojson(self):
        return str(self)

    @classmethod
    def awkward_form(cls, file, context):
        return uproot.containers.AsString(False, typename="TString").awkward_form(
            file, context
        )

    writable = True
    _is_memberwise = False

    def _serialize(self, out, header, name, tobject_flags):
        import uproot.writing._cascade

        where = len(out)
        for x in self._bases:
            x._serialize(out, True, None, tobject_flags)

        out.append(uproot.serialization.string(self))

        if header:
            num_bytes = sum(len(x) for x in out[where:])
            version = 2
            out.insert(where, uproot.serialization.numbytes_version(num_bytes, version))


uproot.classes["TString"] = Model_TString



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines low-level methods for serialization, including :doc:`uproot.serialization.string`,
which prepends string lengths following ROOT's convention, :doc:`uproot.serialization.numbytes_version`,
the opposite of :doc:`uproot.deserialization.numbytes_version`, and :doc:`uproot.serialization.serialize_object_any`,
the opposite of :doc:`uproot.deserialization.read_object_any`.
"""


import struct

import numpy

import uproot.const
import uproot.deserialization


def string(data):
    """
    Converts a Python string into bytes, ready to be written to a file.

    If the string's byte representation (UTF-8) has fewer than 255 bytes, it
    is preceded by a 1-byte length; otherwise, it is preceded by ``b'\xff'`` and a
    4-byte length.
    """
    return bytestring(data.encode(errors="surrogateescape"))


def bytestring(data):
    """
    Converts Python bytes into a length-prefixed bytestring, ready to be written to a file.

    If the string's byte representation (UTF-8) has fewer than 255 bytes, it
    is preceded by a 1-byte length; otherwise, it is preceded by ``b'\xff'`` and a
    4-byte length.
    """
    length = len(data)
    if length < 255:
        return struct.pack(">B%ds" % length, length, data)
    else:
        return struct.pack(">BI%ds" % length, 255, length, data)


def numbytes_version(num_bytes, version):
    """
    Args:
        num_bytes (int): The first value to include in the header.
        version (int): The second value to include in the header.

    Returns a 6-byte bytes object consisting of ``num_bytes`` and ``version``,
    with the appropriate offset (+2) and ``uproot.const.kByteCountMask`` applied
    to ``num_bytes``.

    This function is the opposite of :doc:`uproot.deserialization.numbytes_version`,
    but only generates one case (whereas the deserializer must handle special flags
    in the ``num_bytes``).
    """
    return uproot.deserialization._numbytes_version_1.pack(
        numpy.uint32(num_bytes + 2) | uproot.const.kByteCountMask, version
    )


_serialize_object_any_format1 = struct.Struct(">II")


def _serialize_object_any(out, model, name):
    if model is None:
        out.append(b"\x00\x00\x00\x00")

    else:
        where = len(out)
        model._serialize(out, True, name, numpy.uint32(0x00000000))

        classname = model.classname.encode(errors="surrogateescape") + b"\x00"
        num_bytes = sum(len(x) for x in out[where:]) + len(classname) + 4
        bcnt = numpy.uint32(num_bytes) | uproot.const.kByteCountMask
        tag = uproot.const.kNewClassTag

        out.insert(where, _serialize_object_any_format1.pack(bcnt, tag) + classname)


def serialize_object_any(model, name=None):
    """
    Args:
        model (:doc:`uproot.model.Model`): Object to serialize.
        name (str or None): If not None, overrides the object's name.

    Serializes the object in a form that could be read by :doc:`uproot.deserialization.read_object_any`.
    """
    out = []
    _serialize_object_any(out, model, name)
    return b"".join(out)



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import array
import os

import numpy as np
import pytest

import uproot

ROOT = pytest.importorskip("ROOT")


def test_from_pyroot():
    pyroot_hist = ROOT.TH1F("hist", "title", 10, -5, 5)
    uproot_hist = uproot.from_pyroot(pyroot_hist)
    assert isinstance(uproot_hist, uproot.Model)
    assert isinstance(uproot_hist, uproot.behaviors.TH1.TH1)
    assert isinstance(uproot_hist.values(), np.ndarray)
    assert uproot_hist.values().tolist() == [0.0] * 10
    assert isinstance(uproot_hist.axis().edges(), np.ndarray)
    assert uproot_hist.axis().edges().tolist() == np.linspace(-5, 5, 11).tolist()

    pyroot_vec = ROOT.TLorentzVector(1, 2, 3, 4)
    uproot_vec = uproot.from_pyroot(pyroot_vec)
    assert isinstance(uproot_vec, uproot.Model)
    assert uproot_vec.member("fP").member("fX") == 1
    assert uproot_vec.member("fP").member("fY") == 2
    assert uproot_vec.member("fP").member("fZ") == 3
    assert uproot_vec.member("fE") == 4


def test_write_pyroot_TObjString(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout["something"] = ROOT.TObjString("wowie")

    with uproot.open(newfile) as fin:
        assert fin["something"] == "wowie"

    f = ROOT.TFile(newfile)
    assert str(f.Get("something")) == "wowie"
    f.Close()


def test_write_pyroot_TLorentzVector(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout["something"] = ROOT.TLorentzVector(1, 2, 3, 4)

    classes = dict(uproot.classes)
    with uproot.open(newfile, custom_classes=classes) as fin:
        uproot_vec = fin["something"]
        assert uproot_vec.member("fP").member("fX") == 1
        assert uproot_vec.member("fP").member("fY") == 2
        assert uproot_vec.member("fP").member("fZ") == 3
        assert uproot_vec.member("fE") == 4

    f = ROOT.TFile(newfile)
    pyroot_vec = f.Get("something")
    assert pyroot_vec.X() == 1
    assert pyroot_vec.Y() == 2
    assert pyroot_vec.Z() == 3
    assert pyroot_vec.T() == 4
    f.Close()


def test_write_pyroot_TH1F(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout["something"] = ROOT.TH1F("something", "wicked this way comes", 10, -5, 5)

    with uproot.open(newfile) as fin:
        uproot_hist = fin["something"]
        assert uproot_hist.values().tolist() == [0.0] * 10
        assert uproot_hist.axis().edges().tolist() == np.linspace(-5, 5, 11).tolist()

    f = ROOT.TFile(newfile)
    pyroot_hist = f.Get("something")
    assert [pyroot_hist.GetBinContent(i) for i in range(1, 11)] == [0.0] * 10
    assert pyroot_hist.GetBinLowEdge(1) == pytest.approx(-5)
    assert pyroot_hist.GetBinWidth(1) == pytest.approx(1)
    f.Close()


def test_convert_to_pyroot():
    uproot_tobjstring = uproot.to_writable("hello")
    uproot_histogram = uproot.to_writable(
        (np.array([3, 2, 1.0]), np.array([10, 15, 20, 25.0]))
    )

    pyroot_tobjstring = uproot_tobjstring.to_pyroot()
    pyroot_histogram = uproot_histogram.to_pyroot()

    assert str(pyroot_tobjstring) == "hello"
    assert pyroot_histogram.GetBinContent(1) == 3
    assert pyroot_histogram.GetBinContent(2) == 2
    assert pyroot_histogram.GetBinContent(3) == 1
    assert pyroot_histogram.GetBinLowEdge(1) == pytest.approx(10)
    assert pyroot_histogram.GetBinWidth(1) == pytest.approx(5)



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy
import pytest
import skhep_testdata

import uproot


@pytest.fixture(scope="module")
def datafile(tmpdir_factory):
    yield skhep_testdata.data_path("uproot-issue-240.root")


@pytest.fixture
def graph(datafile):
    with uproot.open(datafile) as f:
        yield f["Expected limit 1lbb/Graph1D_y1"]


def test_interpretation(graph):
    assert graph.classname == "TGraphAsymmErrors"
    assert graph.behaviors[0] == uproot.behaviors.TGraphAsymmErrors.TGraphAsymmErrors

    assert "fX" in graph.all_members.keys()
    assert "fY" in graph.all_members.keys()
    assert "fEXlow" in graph.all_members.keys()
    assert "fEYlow" in graph.all_members.keys()
    assert "fEXhigh" in graph.all_members.keys()
    assert "fEYhigh" in graph.all_members.keys()


@pytest.mark.parametrize("axis", [-2, -1, 0, 1, "x", "y"])
def test_values_single(graph, axis):
    values = graph.values(axis=axis)
    assert isinstance(values, numpy.ndarray)
    assert values.shape == (162,)


@pytest.mark.parametrize("axis", [-2, -1, 0, 1, "x", "y"])
@pytest.mark.parametrize("which", ["low", "high", "mean", "diff"])
def test_errors_single(graph, axis, which):
    errors = graph.errors(axis=axis, which=which)
    assert isinstance(errors, numpy.ndarray)
    assert errors.shape == (162,)


@pytest.mark.parametrize("axis", ["both"])
def test_values_double(graph, axis):
    values = graph.values(axis=axis)
    assert len(values) == 2
    assert all(isinstance(arr, numpy.ndarray) for arr in values)
    assert all(arr.shape == (162,) for arr in values)


@pytest.mark.parametrize("axis", ["both"])
@pytest.mark.parametrize("which", ["low", "high", "mean", "diff"])
def test_errors_double(graph, axis, which):
    errors = graph.errors(axis=axis, which=which)
    assert len(errors) == 2
    assert all(isinstance(arr, numpy.ndarray) for arr in errors)
    assert all(arr.shape == (162,) for arr in errors)



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines the behavior of ``TProfile``.
"""


import numpy

import uproot
from uproot.behaviors.TH1 import boost_axis_metadata, boost_metadata

_kERRORMEAN = 0
_kERRORSPREAD = 1
_kERRORSPREADI = 2
_kERRORSPREADG = 3


# closely follows the ROOT function, using the same names (with 'root_' prepended)
# https://github.com/root-project/root/blob/ffc7c588ac91aca30e75d356ea971129ee6a836a/hist/hist/src/TProfileHelper.h#L141-L163
def _effective_counts_1d(fBinEntries, fBinSumw2, fNcells):
    root_sumOfWeights = fBinEntries
    root_sumOfWeights = numpy.asarray(root_sumOfWeights, dtype=numpy.float64)

    root_sumOfWeightSquare = fBinSumw2
    root_sumOfWeightSquare = numpy.asarray(root_sumOfWeightSquare, dtype=numpy.float64)

    if len(root_sumOfWeightSquare) == 0 or len(root_sumOfWeightSquare) != fNcells:
        return root_sumOfWeights

    positive = root_sumOfWeightSquare > 0

    out = numpy.zeros(len(root_sumOfWeights), dtype=numpy.float64)
    out[positive] = root_sumOfWeights[positive] ** 2 / root_sumOfWeightSquare[positive]
    return out


# duplicates the first part of '_values_errors_1d'
def _values_1d(fBinEntries, root_cont):
    root_sum = fBinEntries
    root_sum = numpy.asarray(root_sum, dtype=numpy.float64)
    nonzero = root_sum != 0

    root_contsum = numpy.zeros(len(root_cont), dtype=numpy.float64)
    root_contsum[nonzero] = root_cont[nonzero] / root_sum[nonzero]

    return root_contsum


def _error_mode_str(error_mode):
    if error_mode is None or error_mode == _kERRORMEAN or error_mode == "":
        return ""
    elif error_mode == _kERRORSPREAD or error_mode == "s":
        return "S"
    elif error_mode == _kERRORSPREADI or error_mode == "i":
        return "I"
    elif error_mode == _kERRORSPREADG or error_mode == "g":
        return "G"
    else:
        return "_"


# closely follows the ROOT function, using the same names (with 'root_' prepended)
# https://github.com/root-project/root/blob/ffc7c588ac91aca30e75d356ea971129ee6a836a/hist/hist/src/TProfileHelper.h#L660-L721
def _values_errors_1d(error_mode, fBinEntries, root_cont, fSumw2, fNcells, fBinSumw2):
    if error_mode is None or error_mode == _kERRORMEAN or error_mode == "":
        error_mode = _kERRORMEAN
    elif error_mode == _kERRORSPREAD or error_mode == "s":
        error_mode = _kERRORSPREAD
    elif error_mode == _kERRORSPREADI or error_mode == "i":
        error_mode = _kERRORSPREADI
    elif error_mode == _kERRORSPREADG or error_mode == "g":
        error_mode = _kERRORSPREADG
    else:
        raise ValueError(
            "error_mode must be None/0/'' for error-on-mean,\n"
            "                        1/'s' for spread (variance),\n"
            "                        2/'i' for integer spread (using sqrt(12)),\n"
            "                     or 3/'g' for Gaussian spread\n"
            "                    not "
            + repr(error_mode)
            + "see https://root.cern.ch/doc/master/classTProfile.html"
        )

    root_sum = fBinEntries
    root_sum = numpy.asarray(root_sum, dtype=numpy.float64)
    nonzero = root_sum != 0

    root_contsum = numpy.zeros(len(root_cont), dtype=numpy.float64)
    root_contsum[nonzero] = root_cont[nonzero] / root_sum[nonzero]

    if error_mode == _kERRORSPREADG:
        out = numpy.zeros(len(root_cont), dtype=numpy.float64)
        out[nonzero] = 1.0 / numpy.sqrt(root_sum[nonzero])
        return root_contsum, out

    root_err2 = fSumw2
    if root_err2 is None or len(root_err2) != fNcells:
        root_err2 = numpy.zeros(len(root_cont), dtype=numpy.float64)
    else:
        root_err2 = numpy.asarray(root_err2, dtype=numpy.float64)

    root_neff = _effective_counts_1d(fBinEntries, fBinSumw2, fNcells)

    root_eprim2 = numpy.zeros(len(root_cont), dtype=numpy.float64)
    root_eprim2[nonzero] = abs(
        root_err2[nonzero] / root_sum[nonzero] - root_contsum[nonzero] ** 2
    )
    root_eprim = numpy.sqrt(root_eprim2)

    if error_mode == _kERRORSPREADI:
        numer = numpy.ones(len(root_cont), dtype=numpy.float64)
        denom = numpy.full(len(root_cont), numpy.sqrt(12 * root_neff))

        eprim_nonzero = root_eprim != 0
        numer[eprim_nonzero] = root_eprim[eprim_nonzero]
        denom[eprim_nonzero] = numpy.sqrt(root_neff[eprim_nonzero])

        out = numpy.zeros(len(root_cont), dtype=numpy.float64)
        out[nonzero] = numer[nonzero] / denom[nonzero]
        return root_contsum, out

    if error_mode == _kERRORSPREAD:
        root_eprim[~nonzero] = 0.0
        return root_contsum, root_eprim

    out = numpy.zeros(len(root_cont), dtype=numpy.float64)
    out[nonzero] = root_eprim[nonzero] / numpy.sqrt(root_neff[nonzero])
    return root_contsum, out


class Profile(uproot.behaviors.TH1.Histogram):
    """
    Abstract class for profile plots.
    """

    @property
    def kind(self):
        return "MEAN"

    def counts(self, flow=False):
        """
        Args:
            flow (bool): If True, include underflow and overflow bins before and
                after the normal (finite-width) bins.

        The effective number of entries, which is a step in the calculation of
        :ref:`uproot.behaviors.TProfile.Profile.values`. The calculation
        of profile errors exactly follows ROOT's "effective entries", but in a
        vectorized NumPy form.
        """
        raise NotImplementedError(repr(self))

    def values(self, flow=False):
        """
        Args:
            flow (bool): If True, include underflow and overflow bins before and
                after the normal (finite-width) bins.

        Mean value of each bin as a 1, 2, or 3 dimensional ``numpy.ndarray`` of
        ``numpy.float64``.

        Setting ``flow=True`` increases the length of each dimension by two.
        """
        raise NotImplementedError(repr(self))

    def errors(self, flow=False, error_mode=""):
        """
        Args:
            flow (bool): If True, include underflow and overflow bins before and
                after the normal (finite-width) bins.
            error_mode (str): Choose a method for calculating the errors (see below).

        Errors (uncertainties) in the :ref:`uproot.behaviors.TH1.Histogram.values`
        as a 1, 2, or 3 dimensional ``numpy.ndarray`` of ``numpy.float64``.

        The calculation of profile errors exactly follows ROOT's function, but
        in a vectorized NumPy form. The ``error_mode`` may be

        * ``""`` for standard error on the mean
        * ``"s"`` for spread
        * ``"i"`` for integer data
        * ``"g"`` for Gaussian

        following `ROOT's profile documentation <https://root.cern.ch/doc/master/classTProfile.html>`__.

        Setting ``flow=True`` increases the length of each dimension by two.
        """
        values, errors = self._values_errors(flow, error_mode)
        return errors

    def variances(self, flow=False, error_mode=""):
        """
        Args:
            flow (bool): If True, include underflow and overflow bins before and
                after the normal (finite-width) bins.
            error_mode (str): Choose a method for calculating the errors (see below).

        Variances (uncertainties squared) in the
        :ref:`uproot.behaviors.TH1.Histogram.values` as a 1, 2, or 3
        dimensional ``numpy.ndarray`` of ``numpy.float64``.

        The calculation of profile variances exactly follows ROOT's function, but
        in a vectorized NumPy form. The ``error_mode`` may be

        * ``""`` for standard error on the mean (squared)
        * ``"s"`` for spread (squared)
        * ``"i"`` for integer data (squared)
        * ``"g"`` for Gaussian (squared)

        following `ROOT's profile documentation <https://root.cern.ch/doc/master/classTProfile.html>`__.

        Setting ``flow=True`` increases the length of each dimension by two.
        """
        values, errors = self._values_errors(flow, error_mode)
        return numpy.square(errors)


class TProfile(Profile):
    """
    Behaviors for one-dimensional profiles: ROOT's ``TProfile``.
    """

    no_inherit = (uproot.behaviors.TH1.TH1,)

    @property
    def axes(self):
        return (self.member("fXaxis"),)

    def axis(self, axis=0):  # default axis for one-dimensional is intentional
        if axis == 0 or axis == -1 or axis == "x":
            return self.member("fXaxis")
        else:
            raise ValueError("axis must be 0 (-1) or 'x' for a TProfile")

    @property
    def weighted(self):
        fBinSumw2 = self.member("fBinSumw2", none_if_missing=True)
        return fBinSumw2 is None or len(fBinSumw2) != len(self.member("fNcells"))

    def counts(self, flow=True):
        out = _effective_counts_1d(
            self.member("fBinEntries"),
            self.member("fBinSumw2"),
            self.member("fNcells"),
        )
        if flow:
            return out
        else:
            return out[1:-1]

    def values(self, flow=False):
        if hasattr(self, "_values"):
            values = self._values
        else:
            (root_cont,) = self.base(uproot.models.TArray.Model_TArray)
            root_cont = numpy.asarray(root_cont, dtype=numpy.float64)
            values = _values_1d(
                self.member("fBinEntries"),
                root_cont,
            )
            self._values = values

        if flow:
            return values
        else:
            return values[1:-1]

    def _values_errors(self, flow, error_mode):
        attr = "_errors" + _error_mode_str(error_mode)
        if hasattr(self, attr):
            values = self._values
            errors = getattr(self, attr)
        else:
            (root_cont,) = self.base(uproot.models.TArray.Model_TArray)
            root_cont = numpy.asarray(root_cont, dtype=numpy.float64)
            fSumw2 = self.member("fSumw2", none_if_missing=True)
            if fSumw2 is not None:
                fSumw2 = numpy.asarray(fSumw2)
            values, errors = _values_errors_1d(
                error_mode,
                numpy.asarray(self.member("fBinEntries")),
                root_cont,
                fSumw2,
                self.member("fNcells"),
                numpy.asarray(self.member("fBinSumw2")),
            )
            self._values = values
            setattr(self, attr, errors)

        if flow:
            return values, errors
        else:
            return values[1:-1], errors[1:-1]

    def to_boost(self, metadata=boost_metadata, axis_metadata=boost_axis_metadata):
        boost_histogram = uproot.extras.boost_histogram()

        effective_counts = self.counts(flow=True)
        values, errors = self._values_errors(True, self.member("fErrorMode"))
        variances = numpy.square(errors)
        sum_of_bin_weights = numpy.asarray(self.member("fBinEntries"))

        storage = boost_histogram.storage.WeightedMean()

        xaxis = uproot.behaviors.TH1._boost_axis(self.member("fXaxis"), axis_metadata)
        out = boost_histogram.Histogram(xaxis, storage=storage)
        for k, v in metadata.items():
            setattr(out, k, self.member(v))

        if isinstance(xaxis, boost_histogram.axis.StrCategory):
            effective_counts = effective_counts[1:]
            values = values[1:]
            variances = variances[1:]
            sum_of_bin_weights = sum_of_bin_weights[1:]

        view = out.view(flow=True)

        # https://github.com/root-project/root/blob/ffc7c588ac91aca30e75d356ea971129ee6a836a/hist/hist/src/TProfileHelper.h#L668-L671
        with numpy.errstate(divide="ignore", invalid="ignore"):
            sum_of_bin_weights_squared = (sum_of_bin_weights**2) / effective_counts

        # TODO: Drop this when boost-histogram has a way to set using the constructor.
        # New version should look something like this:
        # view[...] = np.stack(sum_of_bin_weights, sum_of_bin_weights_squared, values, variances)
        # Current / classic version:
        view["sum_of_weights"] = sum_of_bin_weights
        view["sum_of_weights_squared"] = sum_of_bin_weights_squared
        view["value"] = values
        view["_sum_of_weighted_deltas_squared"] = variances * (
            sum_of_bin_weights - sum_of_bin_weights_squared / sum_of_bin_weights
        )

        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os
import pickle
import sys

import pytest
import skhep_testdata

import uproot

pytest.importorskip("awkward")


def test_pickle_roundtrip_mmap():
    with uproot.open(skhep_testdata.data_path("uproot-small-dy-withoffsets.root")) as f:
        pkl = pickle.dumps(f["tree"])

    branch = pickle.loads(pkl)["Muon_pt"]
    muonpt1 = branch.array(library="np", entry_start=10, entry_stop=20)
    assert [x.tolist() for x in muonpt1] == [
        [20.60145378112793],
        [50.36957550048828, 41.21387481689453, 3.1869382858276367],
        [51.685970306396484, 35.227813720703125],
        [],
        [],
        [],
        [],
        [23.073759078979492],
        [32.921417236328125, 8.922308921813965, 4.368383407592773],
        [51.9132194519043, 31.930095672607422],
    ]


@pytest.mark.network
def test_pickle_roundtrip_http():
    with uproot.open("https://scikit-hep.org/uproot3/examples/Zmumu.root") as f:
        pkl = pickle.dumps(f["events"])

    tree = pickle.loads(pkl)
    assert tree.num_entries == 2304
    assert list(tree["M"].array(entry_stop=10)) == [
        82.4626915551,
        83.6262040052,
        83.3084646667,
        82.1493728809,
        90.4691230355,
        89.7576631707,
        89.7739431721,
        90.4855320463,
        91.7737014813,
        91.9488195857,
    ]


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_pickle_roundtrip_xrootd():
    pytest.importorskip("XRootD")
    with uproot.open(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root"
    ) as f:
        pkl = pickle.dumps(f["Events"])

    tree = pickle.loads(pkl)
    assert tree.num_entries == 29308627
    assert set(tree["run"].array(entry_stop=10)) == {194778}



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import datetime

import numpy
import pytest
import skhep_testdata

import uproot


@pytest.fixture(scope="module")
def datafile(tmpdir_factory):
    yield skhep_testdata.data_path("uproot-issue-407.root")


@pytest.fixture(params=["foo", "foo_padded"])
def _object(request, datafile):
    with uproot.open(datafile) as f:
        yield f[request.param]


@pytest.fixture
def tree(datafile):
    with uproot.open(datafile) as f:
        yield f["tree"]


def test_streamer(_object):
    assert _object.members["d"].to_datetime() == datetime.datetime(2021, 1, 1, 0, 0, 0)


def test_strided_interpretation(tree):
    assert list(tree.iterate(library="np", how=tuple))[0][0][0].member(
        "fDatime"
    ) == uproot._util.datetime_to_code(datetime.datetime(2021, 1, 1, 0, 0, 0))



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy
import pytest
import skhep_testdata

import uproot


def test():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-vectorVectorDouble.root")) as f:
        array = f["t/x"].array(entry_stop=2)
        assert str(awkward.type(array)) == "2 * var * var * float64"



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest
import skhep_testdata

import uproot


@pytest.fixture(scope="module")
def delphes_tree():
    with uproot.open(skhep_testdata.data_path("uproot-delphes-pr442.root")) as f:
        yield f["Delphes"]


def test_read_delphes_np(delphes_tree):
    branch = delphes_tree["Jet/Jet.SoftDroppedP4[5]"]
    nparray = branch.array(library="np")
    assert nparray.shape == (25,)
    assert nparray[0].shape == (2, 5)
    assert nparray[0][0, 0].members["fE"] == 126.46277691787493

    branch = delphes_tree["GenJet04/GenJet04.Constituents"]
    array = branch.array(library="np")
    assert array.shape == (25,)
    assert isinstance(array[0][0], uproot.models.TRef.Model_TRefArray)


def test_read_delphes_ak(delphes_tree):
    awkward = pytest.importorskip("awkward")

    branch = delphes_tree["Jet/Jet.SoftDroppedP4[5]"]
    array = branch.array(library="ak")
    assert array[0, 0, 0].fE == 126.46277691787493
    assert awkward.all(awkward.num(array, axis=2) == 5)

    branch = delphes_tree["GenJet04/GenJet04.Constituents"]
    array = branch.array(library="ak")
    assert set(array.fields) == {"fSize", "fName", "refs"}



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy as np
import skhep_testdata
from numpy.testing import assert_array_equal

import uproot


def test_read_TProfile2D():

    file = skhep_testdata.data_path("uproot-issue-227a.root")

    with uproot.open(file) as h:
        T = h["hprof2d"]

    assert T.kind == "MEAN"
    assert_array_equal(T.axis("x").edges(), np.array([1.0, 2.0, 3.0]))
    assert_array_equal(T.axis("y").edges(), np.array([1.0, 2.0, 3.0, 4.0]))
    assert np.sum(T.counts(flow=True)) == 12
    assert_array_equal(T.values().tolist(), [[1.0, 2.0, 0.0], [2.0, 4.0, 6.0]])


def test_read_TProfile3D():

    file = skhep_testdata.data_path("uproot-issue-227b.root")

    with uproot.open(file) as h:
        T = h["hprof3d"]

    assert T.kind == "MEAN"
    assert_array_equal(T.axis("x").edges(), np.array([1.0, 2.0, 3.0]))
    assert_array_equal(T.axis("y").edges(), np.array([1.0, 2.0, 3.0, 4.0]))
    assert_array_equal(T.axis("z").edges(), np.array([1.0, 2.0, 3.0, 4.0, 5.0]))
    assert np.sum(T.counts(flow=True)) == 12
    assert_array_equal(
        T.values().tolist(),
        [
            [[2.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 6.0, 0.0]],
            [[0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 8.0], [0.0, 0.0, 0.0, 0.0]],
        ],
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys

import numpy
import pytest
import skhep_testdata

import uproot


def test_awkward_strings():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["string"].array(library="ak")) == [
            "one",
            "two",
            "three",
            "four",
            "five",
        ]


def test_pandas_strings():
    pandas = pytest.importorskip("pandas")
    pytest.importorskip("awkward_pandas")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert tree["string"].array(library="pd").values.tolist() == [
            "one",
            "two",
            "three",
            "four",
            "five",
        ]


def test_leaflist_numpy():
    with uproot.open(skhep_testdata.data_path("uproot-leaflist.root"))[
        "tree/leaflist"
    ] as branch:
        result = branch.array(library="np")
        assert result.dtype.names == ("x", "y", "z")
        assert result.tolist() == [
            (1.1, 1, 97),
            (2.2, 2, 98),
            (3.3, 3, 99),
            (4.0, 4, 100),
            (5.5, 5, 101),
        ]
        assert branch.typename == "struct {double x; int32_t y; int8_t z;}"


def test_leaflist_awkward():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-leaflist.root"))[
        "tree/leaflist"
    ] as branch:
        result = branch.array(library="ak")
        assert str(awkward.type(result)) == "5 * {x: float64, y: int32, z: int8}"
        assert awkward.to_list(result) == [
            {"x": 1.1, "y": 1, "z": 97},
            {"x": 2.2, "y": 2, "z": 98},
            {"x": 3.3, "y": 3, "z": 99},
            {"x": 4.0, "y": 4, "z": 100},
            {"x": 5.5, "y": 5, "z": 101},
        ]


def test_leaflist_pandas():
    pandas = pytest.importorskip("pandas")
    pytest.importorskip("awkward_pandas")
    with uproot.open(skhep_testdata.data_path("uproot-leaflist.root"))["tree"] as tree:
        result = tree["leaflist"].array(library="pd")

        assert result.ak["x"].tolist() == [1.1, 2.2, 3.3, 4.0, 5.5]
        assert result.ak["y"].tolist() == [1, 2, 3, 4, 5]
        assert result.ak["z"].tolist() == [97, 98, 99, 100, 101]

        result = tree.arrays("leaflist", library="pd")
        assert list(result.columns) == ["leaflist"]


def test_fixed_width():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as tree:
        assert tree["ai4"].array(library="np").tolist() == [
            [i, i + 1, i + 2] for i in range(-14, 16)
        ]
        assert tree["ai4"].typename == "int32_t[3]"


def test_fixed_width_awkward():
    awkward = pytest.importorskip("awkward")
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as tree:
        assert awkward.to_list(tree["ai4"].array(library="ak")) == [
            [i, i + 1, i + 2] for i in range(-14, 16)
        ]


def test_fixed_width_pandas():
    pandas = pytest.importorskip("pandas")
    pytest.importorskip("awkward_pandas")
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as tree:
        result = tree["ai4"].array(library="pd").ak.array
        assert result[:, 0].tolist() == list(range(-14, 16))
        assert result[:, 1].tolist() == list(range(-13, 17))
        assert result[:, 2].tolist() == list(range(-12, 18))

        result = tree.arrays("ai4", library="pd")
        assert list(result.columns) == ["ai4"]
        assert result["ai4"].ak.array[:, 0].tolist() == list(range(-14, 16))
        assert result["ai4"].ak.array[:, 1].tolist() == list(range(-13, 17))
        assert result["ai4"].ak.array[:, 2].tolist() == list(range(-12, 18))


def test_fixed_width_pandas_2():
    pandas = pytest.importorskip("pandas")
    pytest.importorskip("awkward_pandas")
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree"
    ] as tree:
        result = tree.arrays(
            "xyz", aliases={"xyz": "get('ArrayI32[10]')"}, library="pd"
        )
        assert list(result.columns) == ["xyz"]


def hook(self, **kwargs):
    print("ENTER")
    for k, v in kwargs.items():
        print(k, v)



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest

import uproot


@pytest.mark.network
def test_issue176():
    with uproot.open(
        "https://starterkit.web.cern.ch/starterkit/data/advanced-python-2019/dalitzdata.root"
    ) as f:
        data = f["tree/Y1"].array(library="np")
        assert len(data) == 100000


@pytest.mark.network
def test_issue176_again():
    with uproot.open(
        "https://starterkit.web.cern.ch/starterkit/data/advanced-python-2019/dalitzdata.root"
    ) as f:
        data = f["tree"].arrays(["Y1", "Y2"], library="np")
        assert len(data["Y1"]) == 100000
        assert len(data["Y2"]) == 100000


@pytest.mark.network
def test_issue121():
    with uproot.open(
        "https://github.com/CoffeaTeam/coffea/raw/master/tests/samples/nano_dy.root"
    ) as f:
        data = f["Events/MET_pt"].array(library="np")
        assert len(data) == 40
